import logging
import os
import subprocess
import gevent
import random
from typing import Any, Dict
from pathlib import Path

from openfabric_pysdk.context import Ray, State, StateStatus, RaySchema, RayStatus, StateSchema
from openfabric_pysdk.loader import ConfigClass, InputClass, OutputClass
from openfabric_pysdk.loader.config import manifest, state_config
from openfabric_pysdk.store import KeyValueDB

from .comm import Subscriber, Publisher, DispatchMessage, DispatchActions

my_env = os.environ.copy()
if "PYTHONPATH" in my_env:
    my_env["PYTHONPATH"] = f"{os.getcwd()}:{my_env['PYTHONPATH']}"
else:
    my_env["PYTHONPATH"] = os.getcwd()


#######################################################
#  App
#######################################################
class App:
    state: State = None
    __process = None
    __file_path = "actions.log"
    __script_dir = Path(__file__).parent.absolute()

    # ------------------------------------------------------------------------
    def __init__(self):
        self.state = State()
        self.publisher_port = random.randint(5001, 9999)
        self.subscriber_port = self.publisher_port + 1
        self.publisher = Publisher(f"tcp://127.0.0.1:{self.publisher_port}")
        self.subscriber = Subscriber(f"tcp://127.0.0.1:{self.subscriber_port}", self.__message_received_callback)
        self.dispatch("START", "app")

    # ------------------------------------------------------------------------
    def __del__(self):
        # TODO: could signal 
        #    self.publisher.publish(DispatchMessage.exit("closing"))
        # in order to have a graceful shutdown
        # if no response after a period of time, kill self.__process
        # stop publisher and subscriber
        pass

    # ------------------------------------------------------------------------
    def __message_received_callback(self, message):
        from openfabric_pysdk.engine import engine
        action, data = DispatchMessage.deserialize(message)

        if action is None:
            logging.error("Missing action in message")
            return

        if action == DispatchActions.FETCH:
            if data == "queue":
                logging.debug("FETCH: " + str(data))
                self.publisher.publish(DispatchMessage.state_update(DispatchActions.FETCH, data))
            else:
                logging.error("Unknown fetch request: " + str(data))
        elif action == DispatchActions.UPDATE:
            response: Dict[str, Any] = data

            if 'qid' not in response:
                logging.error("Missing qid in message")
                return

            qid = response["qid"]
            logging.debug("UPDATE: " + str(qid))

            flush_data = False
            if 'output' in response:
                engine.write(qid, "out", response['output'])
                flush_data |= True

            if 'input' in response:
                engine.write(qid, "in", response['input'])
                flush_data |= True

            if 'ray' in response:
                ray = RaySchema().load(response['ray'])
                engine.ray(ray.qid).update(ray)
                engine.write(qid, "ray", response['ray'])
                flush_data |= ray.finished
            else:
                ray = engine.ray(qid)

            if ray is not None and flush_data:
                engine.flush(qid)

        elif action == DispatchActions.APP_STATE:
            receivedState = StateSchema().load(data)
            logging.info(f"State update: {self.state.status} -> {receivedState['status']}")
            self.state.status = receivedState["status"]

        elif action == DispatchActions.LOG:
            logging.log(data['level'], f"Worker-log: {data['message']}")
        elif action == DispatchActions.EXIT:
            logging.info("Worker exited with reason: " + str(data))
            # Temporary workaround
            self.__process.kill()
            self.__process = None
        else:
            logging.error("Unexpected action: " + str(action))

    # ------------------------------------------------------------------------
    def set_status(self, status: StateStatus):
        self.state.status = status
        self.dispatch("STATUS", status, False)

    # ------------------------------------------------------------------------
    def dispatch(self, event, data="", start_worker=True):
        logging.debug(f"Dispatching {event} {data}")

        if start_worker:
            self.__update_worker_state()

        if event == "ADD":
            self.publisher.publish(DispatchMessage.add(data))
        elif event == "DEL":
            self.publisher.publish(DispatchMessage.remove(data))
        elif event == "CFG":
            self.publisher.publish(DispatchMessage.configure())
        elif event == "CHK":
            self.publisher.publish(DispatchMessage.check_request(data))

    # ------------------------------------------------------------------------
    def __update_worker_state(self):
        if self.__process is None or self.__process.poll() is not None:
            self.dispatch("START", "worker", False)
            # Note: remove stdout/stderr redirection to debug app
            # self.__process = subprocess.Popen(["python3", self.__script_dir / "worker.py", "--publisher_port", str(self.subscriber_port), "--subscriber_port", str(self.publisher_port)], env=my_env)
            self.__process = subprocess.Popen(
                ["python3", self.__script_dir / "worker.py", "--publisher_port", str(self.subscriber_port),
                 "--subscriber_port", str(self.publisher_port)], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT,
                env=my_env)

    # ------------------------------------------------------------------------
    def execution_callback_function(self, input: InputClass, ray: Ray) -> OutputClass:
        self.dispatch("ADD", ray.qid)
        counter = 0

        while True:
            if (ray.finished != True):
                gevent.sleep(0.1)
                counter += 1
                # Make sure the worker was not closing while we
                # were proposing a new entry, or that it died.
                if counter % 10 == 0:
                    self.dispatch("CHK", ray.qid)

                # If the app crashed, cancel all ongoing rays.
                if self.state.status == StateStatus.CRASHED:
                    ray.status = RayStatus.FAILED
                    ray.finished = True
                    break
            else:
                break

        return None

    # ------------------------------------------------------------------------
    def cancel_execution(self, ray: Ray):
        ray.status = RayStatus.CANCELED
        ray.complete()
        self.dispatch("DEL", ray.qid, False)

    # ------------------------------------------------------------------------
    def config_callback_function(self, config: Dict[str, ConfigClass]):
        self.dispatch("CFG", "reload", False)

    # ------------------------------------------------------------------------
    def get_manifest(self) -> KeyValueDB:
        return manifest

    # ------------------------------------------------------------------------
    def get_state_config(self) -> KeyValueDB:
        return state_config



import os
import logging
import threading
import traceback
import uuid
from time import sleep
from typing import Any, Dict, List

import traceback

from openfabric_pysdk.benchmark import MeasureBlockTime
from openfabric_pysdk.context import Ray, RaySchema, RayStatus
from openfabric_pysdk.loader import InputSchema, OutputSchema
from openfabric_pysdk.task import Task
from openfabric_pysdk.app import App
from openfabric_pysdk.store import Store


#######################################################
#  Engine
#######################################################
class Engine:
    __app: App = None
    __store: Store = None
    __rays: Dict[str, Ray] = None
    __task: Task = None
    __instances: int = 0
    __running: bool = False
    __current_qid: str = None
    __worker: threading.Thread = None
    __lock: threading.Condition = threading.Condition()

    # ------------------------------------------------------------------------
    def __init__(self):
        self.__rays = dict()
        self.__store = Store(path=f"{os.getcwd()}/datastore")
        self.__lock.acquire()
        if self.__instances == 0:
            self.__task = Task()

            # Populate rays on startup
            for qid in self.__task.all():
                ray = self.read(qid, 'ray', RaySchema().load)
                if ray is None:
                    continue
                self.__rays[qid] = ray

            self.__worker = threading.Thread(target=self.__process, args=())
            self.__worker.start()

        self.__instances = self.__instances + 1
        self.__lock.release()

        # Wait for processing thread to start
        while not self.__running:
            sleep(0.1)

    # ------------------------------------------------------------------------
    def __del__(self):
        self.__lock.acquire()
        if self.__instances > 0:
            self.__lock.release()
            return

        self.__running = False

        self.__lock.notify_all()
        self.__lock.release()

    # ------------------------------------------------------------------------
    def __process(self):
        self.__running = True
        while self.__running:
            self.__lock.acquire()
            self.__current_qid = None
            while self.__running and self.__task.empty():
                self.__lock.wait()
            try:
                self.__current_qid = self.__task.next()
            except:
                logging.warning("Openfabric - queue empty!")
                traceback.print_exc()
            finally:
                self.__lock.release()

            if self.__running and self.__current_qid is not None:
                self.process(self.__current_qid)

    # ------------------------------------------------------------------------
    def prepare(self, app: App, data: str, qid=None, sid=None, uid=None, rid=None) -> str:
        self.__lock.acquire()
        if qid is None:
            qid: str = uuid.uuid4().hex
        ray = self.ray(qid)
        ray.status = RayStatus.QUEUED
        ray.sid = sid
        ray.uid = uid
        ray.rid = rid
        self.write(qid, 'ray', ray, RaySchema().dump)
        self.write(qid, 'in', data)
        # Make sure the worker is able to read the input and state.
        self.flush(qid)
        self.__app = app
        self.__task.add(qid)
        self.__lock.notify_all()
        self.__lock.release()
        return qid

    # ------------------------------------------------------------------------
    def ray(self, qid: str) -> Ray:
        if self.__rays.get(qid) is None:
            ray = Ray(qid=qid)
            self.__rays[qid] = ray
        return self.__rays[qid]

    # ------------------------------------------------------------------------
    def rays(self, criteria=None) -> List[Ray]:
        rays: List[Ray] = []
        for qid, ray in self.__rays.items():
            if criteria is None or criteria(ray):
                rays.append(ray)
        return rays

    # ------------------------------------------------------------------------
    def pending_rays(self, criteria=None) -> List[str]:
        rays: List[str] = []
        for qid, ray in self.__rays.items():
            if criteria is None or criteria(ray):
                rays.append(ray)
        rays.sort(key=lambda r: r.created_at)
        return rays

    # ------------------------------------------------------------------------
    def process(self, qid):

        if self.__app is None:
            logging.error("Openfabric - no app configured!")
            traceback.print_exc()
            return

        with MeasureBlockTime("Engine::execution_callback_function"):
            ray = self.ray(qid)
            self.__app.execution_callback_function(None, ray)
        output = self.read(qid, 'out', OutputSchema().load)

        return output

    # ------------------------------------------------------------------------
    def delete(self, qid: str, app: App = None) -> Ray:
        self.__lock.acquire()
        self.__task.rem(qid)
        ray = self.ray(qid)
        if app is not None:
            app.cancel_execution(ray)
        else:
            self.__app.cancel_execution(ray)
        self.__store.drop(qid)
        self.__rays.pop(qid)
        ray.status = RayStatus.REMOVED
        self.__lock.notify_all()
        self.__lock.release()
        return ray

    # ------------------------------------------------------------------------
    def read(self, qid: str, key: str, deserializer=None) -> Any:
        output = self.__store.get(qid, key)
        if output is None:
            return None
        output = deserializer(output) if deserializer is not None else output
        return output

    # ------------------------------------------------------------------------
    def write(self, qid: str, key: str, val: Any, serializer=None):
        if val is not None:
            val = serializer(val) if serializer is not None else val
        self.__store.set(qid, key, val)

    def flush(self, qid: str):
        self.__store.flush(qid)


engine = Engine()



import gevent
import pickle
import threading
import zmq

from openfabric_pysdk.context import State


class DispatchActions:
    ADD = 'add'
    CHECK = 'check'
    CONFIGURE = 'configure'
    EXIT = 'exit'
    FETCH = 'fetch'
    LOG = 'log'
    REMOVE = 'remove'
    APP_STATE = 'app_state'
    UPDATE = 'update'


class DispatchMessage:
    def __init__(self):
        pass

    @staticmethod
    def add(qid: str):
        return pickle.dumps({'action': DispatchActions.ADD, 'data': qid})

    @staticmethod
    def check_request(qid: str):
        return pickle.dumps({'action': DispatchActions.CHECK, 'data': qid})

    @staticmethod
    def configure():
        return pickle.dumps({'action': DispatchActions.CONFIGURE})

    @staticmethod
    def exit(reason: str):
        return pickle.dumps({'action': DispatchActions.EXIT, 'data': reason})

    @staticmethod
    def fetch(field: str):
        return pickle.dumps({'action': DispatchActions.FETCH, 'data': field})

    @staticmethod
    def log(level: int, message: str):
        data = {'level': level, 'message': message}
        return pickle.dumps({'action': DispatchActions.LOG, 'data': data})

    @staticmethod
    def remove(qid: str):
        return pickle.dumps({'action': DispatchActions.REMOVE, 'data': qid})

    @staticmethod
    def state_update(qid: str, input=None, output=None, ray=None):
        data = {'qid': qid}
        if input is not None:
            data['input'] = input
        if output is not None:
            data['output'] = output
        if ray is not None:
            data['ray'] = ray

        return pickle.dumps({'action': DispatchActions.UPDATE, 'data': data})

    @staticmethod
    def app_state(state: str):
        return pickle.dumps({'action': DispatchActions.APP_STATE, 'data': state})

    @staticmethod
    def deserialize(input):
        data = pickle.loads(input)

        return data['action'], data['data'] if 'data' in data else None


class Publisher:
    def __init__(self, address="tcp://127.0.0.1:5556"):
        context = zmq.Context()
        self.publisher_socket = context.socket(zmq.PUB)
        self.publisher_socket.bind(address)

    def publish(self, message):
        self.publisher_socket.send(message)

    def close(self):
        self.publisher_socket.close()


class Subscriber():
    def __init__(self, address="tcp://127.0.0.1:5556", callback=None, geventEnv=False):
        context = zmq.Context()

        self.subscriber_socket = context.socket(zmq.SUB)
        self.subscriber_socket.connect(address)
        self.subscriber_socket.setsockopt_string(zmq.SUBSCRIBE, "")

        self.running = True

        self.on_message_received_callback = callback
        if geventEnv:
            self.subscriber_greenlet = gevent.spawn(self.__message_handler2, self.subscriber_socket)
        else:
            self.__execution = threading.Thread(target=self.__message_handler, args=(self.subscriber_socket,))
            self.__execution.start()

    def __del__(self):
        pass
        self.running = False
        self.subscriber_socket.close()
        if self.__execution is not None:
            self.__execution.join()
        self.__execution = None

    def __message_handler(self, socket):
        while True:
            try:
                message = socket.recv(flags=zmq.NOBLOCK)
                if self.on_message_received_callback is not None:
                    self.on_message_received_callback(message)
            except zmq.Again:
                gevent.sleep(0.1)
                pass

    # Todo: try to switch to this message handler
    def __message_handler2(self, socket):
        while True:
            message = socket.recv()
            if self.on_message_received_callback is not None:
                self.on_message_received_callback(message)
            gevent.sleep(0.001)

    def register_callback(self, callback):
        self.on_message_received_callback = callback

    def close(self):
        self.subscriber_socket.close()



#!/usr/bin/env python3
import logging
import os
import traceback

import argparse
import re
import time
import threading
import traceback
from typing import Any

import sys

sys.path.append(str({os.getcwd()}))
from comm import Subscriber, Publisher, DispatchMessage, DispatchActions

from openfabric_pysdk.benchmark import MeasureBlockTime
from openfabric_pysdk.context import MessageType, Ray, RaySchema, RayStatus, State, StateStatus, StateSchema
from openfabric_pysdk.loader import InputSchema, OutputSchema
from openfabric_pysdk.store import KeyValueDB
from openfabric_pysdk.loader.config import state_config
from openfabric_pysdk.loader import ConfigSchema

config_callback_function = None
execution_callback_function = None
suspend_callback_function = None

suspend_request_time_s = 5
loglevel = logging.DEBUG
current_qid = None
publisher = None
subscriber = None
worker = None
running = True
working = True
queue = []
lock: threading.Condition = threading.Condition()


def notify_ray_update(ray: Ray):
    logging.debug(f'notify_ray_update {ray}')
    publisher.publish(DispatchMessage.state_update(ray.qid, ray=RaySchema().dump(ray)))
    pass


class LogsHandler(logging.StreamHandler):
    level = loglevel

    def emit(self, record=None):
        if record is None:
            return
        try:
            msg = self.format(record)

            publisher.publish(DispatchMessage.log(record.levelno, msg))
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


class Worker:
    __db__path: KeyValueDB = f"{os.getcwd()}/datastore"

    def __init__(self):
        self.state = State()

    # ------------------------------------------------------------------------
    def process(self, qid):

        with MeasureBlockTime("Engine::execution_callback_function"):
            kvdb = KeyValueDB(f"{qid}", path=self.__db__path, autodump=False)
            output = None
            ray = self.read(kvdb, 'ray', RaySchema().load)
            if ray is None:
                return

            if ray.finished == True and ray.status != RayStatus.REMOVED:
                output = self.read(kvdb, 'out', OutputSchema().load)
                publisher.publish(DispatchMessage.state_update(qid, output=OutputSchema().dump(output)))
                return

            ray.on_update(notify_ray_update)

            logging.info(f'processing {qid}')
            try:
                data = self.read(kvdb, 'in', InputSchema().load)
                if data is None:
                    return

                ray.status = RayStatus.RUNNING

                # Callback execution method
                output = execution_callback_function(data, ray, self.state)

                ray.status = RayStatus.COMPLETED
            except:
                error = f"process - failed executing: [{qid}]\n{traceback.format_exc()}"
                logging.error(error)
                ray.message(MessageType.ERROR, error)
                ray.status = RayStatus.FAILED
        publisher.publish(DispatchMessage.state_update(qid, output=OutputSchema().dump(output)))
        ray.complete()

        logging.info(f'completed {qid}')

    # ------------------------------------------------------------------------
    def read(self, kvdb: KeyValueDB, key: str, deserializer=None) -> Any:
        output = kvdb.get(key)
        if output is None:
            return None
        output = deserializer(output) if deserializer is not None else output
        return output


def process_line(line: str, skipConfig=False):
    global current_qid
    global queue
    global lock
    logging.debug(f'event {line}')

    pattern = re.compile(r'([0-9]+-[0-9]+-[0-9]+\/[0-9]+:[0-9]+) ([0-9a-zA-Z]+)[ ]+(.*)\n')
    event = re.search(pattern, line)

    if (event != None):
        date = event.group(1)
        action = event.group(2)
        data = event.group(3)
        logging.debug(f"Event: date: {date} action: {action} data: {data}")

        if action == "ADD":
            lock.acquire()
            if data != current_qid and data not in queue:
                queue.append(data)
            lock.release()
        elif action == "DEL":
            # TODO: Canceling an execution does not stop it if already started.
            # Status will be sent to the application after execution is completed.
            lock.acquire()
            try:
                while data in queue:
                    queue.remove(data)
            except:
                pass
            lock.release()
        elif action == "CFG":
            if not skipConfig:
                state = State()

                state_config.reload()
                items = state_config.all().items()
                config = dict(map(lambda kv: (kv[0], ConfigSchema().load(kv[1])), items))

                if config_callback_function is not None:
                    try:
                        config_callback_function(config, state)
                    except Exception as e:
                        logging.error(f"Openfabric - invalid configuration can\'t restored : {e}")
                else:
                    logging.warning(f"Openfabric - no configuration callback available")
        elif action == "START":
            logging.debug(f"Start event")
        elif action == "STATUS":
            logging.debug(f"App status {data}")
        else:
            logging.warn(f"Unknown action: {action}")
    else:
        logging.warn(f"Unknown event: {line}")


# would be the main thread
def execution():
    global suspend_request_time_s
    global current_qid
    global running
    global working
    global queue
    global lock
    retry_counter = 10 * suspend_request_time_s

    # TODO: use locks
    while running:
        lock.acquire()
        current_qid = None
        if len(queue) > 0:
            current_qid = queue.pop(0)
        lock.release()
        if current_qid is not None:
            working = True
            try:
                worker.process(current_qid)
            except:
                # TODO: fix this.
                # try catch should be in process, but it's relating to reading the db
                logging.error(f"Failed to execute {current_qid}")
                pass
            retry_counter = 10 * suspend_request_time_s
        else:
            if retry_counter > 0:
                retry_counter -= 1
            else:
                working = False

                # In practice we could implement a suspend_prepare and a suspend function
                # The suspend_prepare would be called before the suspend function once
                # the queue is empty.
                # After a predefined/fixed period we call the suspend function.
                # But the application could return False on the current suspend on the first
                # few calls and True on a later call. Basically the application could decide
                # to suspend, after a predefined period.
                if suspend_callback_function is not None:
                    try:
                        state = State()
                        # If the suspend function returns True, we exit the application.
                        if suspend_callback_function(state):
                            logging.info(f"Openfabric - suspend allowed by the application")
                            publisher.publish(DispatchMessage.exit("suspend"))
                            running = False
                            continue
                        else:
                            logging.debug(f"Openfabric - suspend denied by the application - retrying in 1 second")
                            # retry after 1 second
                            retry_counter = 10
                    except Exception as e:
                        logging.error(f"Openfabric - suspend ended with error : {e}")
                else:
                    # No point in retrying if there is no suspend callback
                    retry_counter = 99999999999

        time.sleep(0.1)


def message_received_callback(message):
    global current_qid
    global queue
    global lock
    global running
    action, data = DispatchMessage.deserialize(message)

    if action == None:
        logging.error("Missing action in message")
        return

    if action == DispatchActions.ADD or action == DispatchActions.CHECK:
        logging.debug(str(action) + ": " + str(data))
        lock.acquire()
        if data != current_qid and data not in queue:
            queue.append(data)
        lock.release()
    elif action == DispatchActions.REMOVE:
        logging.debug("REMOVE: " + str(data))
        lock.acquire()
        try:
            while data in queue:
                queue.remove(data)
        except:
            pass
        lock.release()
    elif action == DispatchActions.CONFIGURE:
        print("CONFIGURE: " + str(data))
        state = State()

        state_config.reload()
        items = state_config.all().items()
        config = dict(map(lambda kv: (kv[0], ConfigSchema().load(kv[1])), items))

        config_callback_function(config, state)
    elif action == DispatchActions.EXIT:
        running = False
    else:
        logging.error("Unexpected action: " + str(action))


if __name__ == "__main__":
    # Disable socketio and engineio logging
    # Reason: would cause recursion error as we are using the proxy to forward logs
    logging.getLogger("socketio").setLevel(logging.ERROR)
    logging.getLogger("engineio").setLevel(logging.ERROR)

    parser = argparse.ArgumentParser(description='Worker')
    parser.add_argument('--debug', action='store_true', help='debug mode')
    parser.add_argument('--publisher_port', type=int, default=5556, help='publisher port')
    parser.add_argument('--subscriber_port', type=int, default=5555, help='subscriber port')
    args = parser.parse_args()

    if args.debug:
        loglevel = logging.DEBUG
    else:
        loglevel = logging.INFO

    # Configure custom logging
    logger = logging.getLogger()
    logger.setLevel(loglevel)
    custom_handler = LogsHandler()
    formatter = logging.Formatter("%(asctime)s: %(message)s")
    custom_handler.setFormatter(formatter)
    logger.addHandler(custom_handler)

    publisher = Publisher(f"tcp://127.0.0.1:{args.publisher_port}")
    subscriber = Subscriber(f"tcp://127.0.0.1:{args.subscriber_port}", message_received_callback)

    try:
        from openfabric_pysdk.api import config_callback_function, \
            execution_callback_function, \
            suspend_callback_function

        startup_error = None

        state = State()
        state.status = StateStatus.RUNNING
        publisher.publish(DispatchMessage.app_state(StateSchema().dump(state)))
    except Exception as inst:
        startup_error = ''.join(traceback.TracebackException.from_exception(inst).format())
        time.sleep(1)
        logging.error(f'Exception while starting app: {startup_error}')
        state = State()
        state.status = StateStatus.CRASHED
        publisher.publish(DispatchMessage.app_state(StateSchema().dump(state)))
        time.sleep(1)
        publisher = None
        subscriber = None
        exit(-1)

    worker = Worker()

    logging.info(f'started')

    exec = threading.Thread(target=execution)
    exec.start()

    while running:
        time.sleep(0.1)

    logging.info(f'exiting')
    exec.join()

    logger.removeHandler(custom_handler)

    # Need to close socket connection otherwise the app hands in state "sleeping"
    worker = None
    logging.info(f'stopped')
    exit(0)



from datetime import datetime
from enum import Enum
from typing import Dict, List

from tqdm.asyncio import tqdm, tqdm_asyncio

from .bar import Bar
from .message import Message, MessageType


#######################################################
#  Ray Status
#######################################################
class RayStatus(Enum):
    QUEUED = 'queued',
    PENDING = 'pending',
    COMPLETED = 'completed',
    RUNNING = 'running',
    CANCELED = 'canceled',
    REMOVED = 'removed',
    UNKNOWN = 'unknown',
    FAILED = 'failed'

    # ------------------------------------------------------------------------
    def __str__(self):
        return self.name


#######################################################
#  Ray
#######################################################
class Ray:

    # ------------------------------------------------------------------------
    def __init__(self, qid):
       # Initialize the attributes through this method to avoid triggering __setattr__
        object.__setattr__(self, 'qid', qid)
        object.__setattr__(self, 'sid', None)
        object.__setattr__(self, 'uid', None)
        object.__setattr__(self, 'rid', None)
        object.__setattr__(self, 'finished', False)
        object.__setattr__(self, 'bars', dict(default=Bar()))
        object.__setattr__(self, 'status', RayStatus.UNKNOWN)
        object.__setattr__(self, 'created_at', datetime.now())
        object.__setattr__(self, 'updated_at', datetime.now())
        object.__setattr__(self, 'messages', list())
        object.__setattr__(self, 'onchange_callback', None)
        object.__setattr__(self, 'tqdms_info', dict())
    
    # ------------------------------------------------------------------------
    def on_update(self, callback):
        object.__setattr__(self, 'onchange_callback', callback)

    # ------------------------------------------------------------------------
    def update(self, other):
        for name, value in vars(other).items():
            if name in ["qid", "sid", "uid", "rid", "onchange_callback"]:
                continue
            object.__setattr__(self, name, value)
        self.__trigger_update()

    # ------------------------------------------------------------------------
    def complete(self, name='default'):
        bar = self.bars.get(name, Bar())
        self.bars[name] = bar
        bar.remaining = 0
        bar.percent = 100
        object.__setattr__(self, "finished", True)
        self.__trigger_update()

    # ------------------------------------------------------------------------
    def progress(self, name='default', step=1, total=100) -> tqdm_asyncio:
        if self.tqdms_info.get(name) is None:
            self.tqdms_info[name] = tqdm(total=total)
        tqdm_bar = self.tqdms_info[name]
        tqdm_bar.update(step)
        # --
        bar = self.bars.get(name, Bar())
        self.bars[name] = bar
        f_dict = tqdm_bar.format_dict
        rate = f_dict.get("rate")
        total = tqdm_bar.total
        n = tqdm_bar.n
        remaining = (total - n) / rate if rate and total else 0
        bar.remaining = max(0, remaining)
        bar.percent = n
        # --
        self.__trigger_update()
        return tqdm_bar

    # ------------------------------------------------------------------------
    def message(self, message_type: MessageType, content: str):
        message = Message()
        message.type = message_type
        message.content = content
        self.messages.append(message)
        self.__trigger_update()

    # ------------------------------------------------------------------------
    def clear_messages(self):
        if self.messages is not None and len(self.messages) > 0:
            self.messages.clear()
            self.__trigger_update()

    # ------------------------------------------------------------------------
    def tqdms(self):
        return self.tqdms_info
    
    def __str__(self):
        return f"Ray(qid={self.qid}, rid={self.rid}, uid={self.uid}, sid={self.sid}, " + \
                 f"status={self.status}, created_at={self.created_at}, updated_at={self.updated_at}, " + \
                 f"finished={self.finished}, messages=[{self.messages}]"

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if key not in ['onchange_callback']:
            self.__trigger_update()

    def __trigger_update(self):
        object.__setattr__(self, "updated_at", datetime.now())
        if self.onchange_callback is not None:
            self.onchange_callback(self)



# ----------------------------
# !!!  Perform GEVENT monkey patch
# ----------------------------
from gevent import monkey
monkey.patch_all()

try:
  # This is in order to be able to use a newer werkzeug version
  # without having to update to the latest flask
  # as the marketplace is still using an older socketIO client
  import werkzeug.serving as werkzeug_serving
  if not hasattr(werkzeug_serving, 'run_with_reloader'):
    import werkzeug
    import werkzeug._reloader
    werkzeug.serving.run_with_reloader = werkzeug._reloader
except:
  # Failed patching werkzeug.
  # Maybe something else is used
  pass

import pathlib
from flask import Flask as Webserver, render_template, request, session
from flask_cors import CORS
from requests import get, post, patch, delete, Request, Response

webserver = Webserver(__name__, template_folder=f"{pathlib.Path(__file__).parent}/../../view")
CORS(webserver)



#!/usr/bin/env python3

import collections
import json
import logging
import threading
import uuid
import zlib
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from typing import Any, Dict

import socketio
import time


# forward declaration
class Proxy: pass


class ExecutionResult:
    def __init__(self, proxy: Proxy):
        self.__progress = None
        self.__proxy = proxy
        self.__rid = uuid.uuid4().hex
        self.__finished = False
        self.last_update = datetime.now()
        self.__result = None
        logging.info(f"Created rid" + str(self.__rid))

    def __del__(self):
        self.cancel()

    def on_progress(self, data):
        self.__progress = data
        self.last_update = datetime.now()

    def on_response(self, data):
        if data != None and "ray" in data:
            self.last_update = datetime.now()
            self.__progress = data["ray"]

        if data != None and "output" in data:
            self.__result = data["output"]

        self.__finished = True
        self.cancel()

    def on_restore(self, data):
        if data != None and "ray" in data:
            self.last_update = datetime.now()
            self.__progress = data["ray"]

        if data != None and "output" in data:
            # TODO: if output != None: force finish?
            self.__result = data["output"]
            self.__finished = True
            self.cancel()

    def request_id(self):
        return self.__rid

    def request_qid(self):
        if self.__progress is not None and \
                "qid" in self.__progress:
            return self.__progress["qid"]
        return None

    def messages(self):
        if self.__progress != None and "messages" in self.__progress:
            return self.__progress["messages"]

        return []

    def progress(self, name="default"):
        if self.__progress != None and "bars" in self.__progress:
            if name in self.__progress["bars"]:
                return self.__progress["bars"][name]

        return None

    def status(self):
        if self.__progress != None and "status" in self.__progress:
            return self.__progress["status"]

        return "UNKNOWN"

    def cancel(self):
        if self.__proxy != None:
            self.__proxy.cancel(self)
            self.__proxy = None

        if self.__finished == False:
            if self.__progress == None:
                self.__progress = {}
            self.__progress["status"] = "CANCELLED"

    def wait(self, timeout=0):
        remaining = timeout

        while self.__proxy != None and (timeout == 0 or remaining > 0) and self.__finished == False:
            remaining -= 0.1
            time.sleep(0.1)

        return self.__finished

    def data(self):
        return self.__result

    def __repr__(self):
        return f"ExecutionResult({self.__rid}, {self.__progress}, {self.__result})"

    def __str__(self):
        return self.__repr__()


#######################################################
#  Proxy
#######################################################
class Proxy:
    # ------------------------------------------------------------------------
    def __init__(self, url: str, tag: str = None):
        self.maximum_no_respose_interval = 15  # seconds
        self.minimum_check_interval = 5  # seconds
        self.minimum_update_interval = 5  # seconds
        self.__results = {}
        self.__executions = {}
        self.__url = url
        self.__tag = url if tag == None else tag;
        self.__cancel_next = False
        self.__pending_responses = 0
        self.__running = True
        self.__sio = socketio.Client()
        self.__lock: threading.RLock = threading.RLock()
        self.__runner = threading.Thread(target=self.__run, args=())
        self.__runner.start()
        self.__checker = threading.Thread(target=self.__check_state, args=())
        self.__checker.start()
        self.__executor = ThreadPoolExecutor(max_workers=1)
        self.__callbacks = collections.defaultdict(list)
        self.__setup_call_backs()

        # 3 seconds timeout
        maxRetries = 30
        while not self.is_connected() and maxRetries > 0:
            time.sleep(0.1)
            maxRetries -= 1

    def __del__(self):
        self.cleanup()

    def disconnect(self):
        self.cleanup()

    def cleanup(self):
        if hasattr(self, '_Proxy__lock'):
            self.__lock.acquire()
        if hasattr(self, '_Proxy__executions'):
            self.__executions.clear()
        if hasattr(self, '_Proxy__lock'):
            self.__lock.release()
        if hasattr(self, '_Proxy__running'):
            self.__running = False
        if hasattr(self, '_Proxy__executor'):
            self.__executor.shutdown(wait=False)
        if hasattr(self, '_Proxy__sio'):
            self.__sio.disconnect()

    def __run(self):
        logging.info(f"{self.__tag}: Proxy started")
        # Wait until the connection with the server ends.
        while self.__running:
            try:
                self.__sio.connect(self.__url, transports='websocket', namespaces=["/app"])
                self.__sio.wait()
                break;
            except:
                logging.exception(f"{self.__tag}: Could not connect to server. Retrying in 15 seconds to {self.__url}")
                maxRetries = 150
                while self.__running and maxRetries > 0:
                    time.sleep(0.1)
                    maxRetries -= 1
                pass

        logging.info(f"{self.__tag}: Proxy exiting")

    def __check_state(self):
        last_update_check = datetime.now()

        logging.info(f"{self.__tag}: State checker started")
        while self.__running:
            # time.sleep(minimum_check_interval)
            while (datetime.now() - last_update_check).total_seconds() < self.minimum_check_interval and self.__running:
                time.sleep(0.1)

            last_update_check = datetime.now()

            executions_to_cancel = []
            executions_to_ping = []
            self.__lock.acquire()
            try:
                for rid, execution in self.__executions.items():
                    elapsedSinceLastUpdate = (datetime.now() - execution.last_update).total_seconds()

                    # TODO: check reference count of the execution so see if it is still in use, and cancel it otherwise
                    # import sys
                    # if sys.getrefcount(execution) < ??:
                    #     executions_to_cancel.append(rid)
                    #     continue

                    if elapsedSinceLastUpdate > self.maximum_no_respose_interval:
                        executions_to_cancel.append(rid)
                    elif elapsedSinceLastUpdate > self.minimum_update_interval:
                        executions_to_ping.append(rid)
            except:
                logging.exception(f"{self.__tag}: Exception in state checker")
                pass
            self.__lock.release()

            try:
                for rid in executions_to_cancel:
                    logging.error(f"{self.__tag}: Execution {rid} is not responding. Cancelling")
                    qid = self.__executions[rid].request_qid()
                    self.__executions[rid].cancel()
                    if qid != None:
                        self.delete(qid)

                for rid in executions_to_ping:
                    qid = self.__executions[rid].request_qid()
                    if qid != None:
                        logging.info(f"{self.__tag}: Execution {rid} is not responding. Requesting update")
                        self.restore(qid)
                    else:
                        logging.info(f"{self.__tag}: Execution {rid} is not responding. Queue id not yet available")
            except:
                logging.exception(f"{self.__tag}: Exception in state checker")
                pass

        logging.info(f"{self.__tag}: State checker exiting")

    def request(self, input, uid):
        access = True
        if (uid == None):
            uid = uuid.uuid4().hex

        execution = ExecutionResult(self)
        rid = execution.request_id()

        # TODO: emit is not threadsafe. We need to lock here.
        #       need also to see if we catch exceptions here.
        data = {}
        data["body"] = input
        data["header"] = {"uid": uid, "rid": rid}
        data = zlib.compress(json.dumps(data).encode('utf-8')), access
        self.__sio.emit('execute', data=data, namespace='/app')

        # Add execution to monitored executions. Assuming rid is unique.
        self.__lock.acquire()
        self.__executions[rid] = execution
        self.__lock.release()

        return execution

    def cancel(self, execution: ExecutionResult):
        self.__lock.acquire()
        rid = execution.request_id()
        if rid in self.__executions:
            self.__executions.pop(rid)
        self.__lock.release()

    # TODO: remove
    def execute_async(self, input, uid=None, rid=None):
        access = True
        if (uid == None):
            uid = uuid.uuid4().hex
        if (rid == None):
            rid = uuid.uuid4().hex
        data = {}
        data["body"] = input
        data["header"] = {"uid": uid, "rid": rid}
        data = zlib.compress(json.dumps(data).encode('utf-8')), access
        self.__sio.emit('execute', data=data, namespace='/app')

        ++self.__pending_responses
        return self.__executor.submit(Proxy.get_response, self, rid)

    # TODO: remove
    def execute(self, input, uid=None, rid=None):
        return self.execute_async(input, uid=uid, rid=rid).result()

    def configure(self, config, uid):
        data = {}
        data["body"] = config
        data["header"] = {"uid": uid}
        data = zlib.compress(json.dumps(data).encode('utf-8'))
        self.__sio.emit('configure', data=data, namespace='/app')

    def is_connected(self):
        return self.__sio.connected

    def is_pending(self):
        return self.__pending_responses > 0

    def cancel_next(self):
        self.__cancel_next = True

    def get_response(self, rid: str):
        ++self.__pending_responses
        return self.__get_response(rid)

    # TODO: remove
    # TODO: should probably return also the error status
    def __get_response(self, rid: str):
        result = None
        print("Check state of rid: ", rid)
        while rid not in self.__results and not self.__cancel_next:
            # TODO: Given that we use a limited number of executors
            # per proxy, we should define a timeout.
            time.sleep(0.1)

        if self.__cancel_next:
            print("Canceled rid: ", rid)
        else:
            print("Got response for rid: ", rid)
            result = self.__results.pop(rid)

        self.__cancel_next = False
        --self.__pending_responses

        return result

    def register(self, action, callback, context=None):
        self.__callbacks[action].append((callback, context))

    def unregister(self, action, callback):
        self.__callbacks[action] = [(f, c) for f, c in self.callbacks[action] if f != callback]

    def __notify(self, action, *args, **kwargs):
        for callback, context in self.__callbacks[action]:
            if context is not None:
                callback(context, *args, **kwargs)
            else:
                callback(*args, **kwargs)

    def resume(self, uid: str):
        self.__sio.emit('resume', data=uid, namespace='/app')

    def restore(self, qid: str):
        self.__sio.emit('restore', data=qid, namespace='/app')

    def state(self, uid: str):
        self.__sio.emit('state', data=uid, namespace='/app')

    def delete(self, qid: str):
        self.__sio.emit('delete', data=qid, namespace='/app')

    def __setup_call_backs(self):
        @self.__sio.event
        def connect():

            logging.info(f"{self.__tag}: Connection established")
            self.__notify("connected", True)

        @self.__sio.event
        def connect_error(data):
            logging.error(f"{self.__tag}: Failed to establish connection")

        @self.__sio.event(namespace='/app')
        def response(data):
            response: Dict[str, Any] = data
            rid = None
            if "ray" in response:
                if "rid" in response["ray"]:
                    rid = response["ray"]["rid"]
                    # TODO: would be good to store results only for known/recent requests
                    #       in order to avoid a case with faulty server sending bad data
                    if "output" in response:
                        self.__results[rid] = (response["output"])

            self.__lock.acquire()
            if rid is not None and rid in self.__executions:
                self.__executions[rid].on_response(response)
            self.__lock.release()

            self.__notify("response", data)

        @self.__sio.event(namespace='/app')
        def submitted(data):
            logging.debug(f"{self.__tag}: Data Received [submitted]: {data}")

            response: Dict[str, Any] = data
            rid = response["rid"]

            self.__lock.acquire()
            if rid in self.__executions:
                self.__executions[rid].on_progress(response)
            self.__lock.release()

            self.__notify("submitted", data)

        @self.__sio.event(namespace='/app')
        def settings(config):
            logging.debug(f"{self.__tag}: Data Received [settings]: {config}")
            response: Dict[str, Any] = config

            self.__notify("settings", response)

        # TODO: move functionality from sample app to here (without the remapping part)
        @self.__sio.event(namespace='/app')
        def progress(data):
            logging.debug(f"{self.__tag}: Data Received [progress]: {data}")

            response: Dict[str, Any] = data
            rid = response["rid"]

            self.__lock.acquire()
            if rid in self.__executions:
                self.__executions[rid].on_progress(response)
            self.__lock.release()

            self.__notify("progress", data)

        @self.__sio.event(namespace='/app')
        def restore(data):
            logging.debug(f"{self.__tag}: Data Received [restore]: {data}")
            response: Dict[str, Any] = data
            rid = None
            if response is not None and \
                    "ray" in response and response["ray"] is not None and \
                    "rid" in response["ray"]:
                rid = response["ray"]["rid"]
            else:
                return

            self.__lock.acquire()
            if rid in self.__executions:
                self.__executions[rid].on_restore(response)
            self.__lock.release()

            self.__notify("restore", data)

        @self.__sio.event(namespace='/app')
        def state(data):
            logging.debug(f"{self.__tag}: Data Received [state]: {data}")
            self.__notify("state", data)

        @self.__sio.event
        def disconnect():
            logging.info(f"{self.__tag}: Disconnected from server")
            self.__notify("connected", False)


# ==============================================================================

def handler(data, *args, **kwargs):
    # logging.info(f"Handler called with data: {data}\n{args}\n------")
    pass


def handler2(data, *args, **kwargs):
    time.sleep(1000)


if __name__ == "__main__":
    # Enable logging
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    # Sample proxy
    wrapper = Proxy('wss://0ff5d32a2fc74601b0781dfd2fc47974.openfabric.network', "App")

    '''
    wrapper.register("response", handler, "response")
    wrapper.register("progress", handler, "progress")
    wrapper.register("submitted", handler, "submitted")
    wrapper.register("restore", handler, "restore")
    wrapper.register("state", handler, "state")
    '''

    input = {"text": ["string"]}

    logging.info("# execution = wrapper.request(input, uid=user)")
    execution = wrapper.request(input, uid="user")

    logging.info("---- > request_id " + str(execution.request_id()))
    logging.info("---- > progress " + str(execution.progress()))
    logging.info("---- > status " + str(execution.status()))
    logging.info("---- > result " + str(execution.data()))
    logging.info("---- > messages " + str(execution.messages()))

    logging.info("# execution.wait()")
    execution.wait()

    logging.info("---- > request_id " + str(execution.request_id()))
    logging.info("---- > progress " + str(execution.progress()))
    logging.info("---- > status " + str(execution.status()))
    logging.info("---- > result " + str(execution.data()))
    logging.info("---- > messages " + str(execution.messages()))

    logging.info("# execution = wrapper.request(input, uid=user)")
    execution = wrapper.request(input, uid="user")
    logging.info("# execution.cancel()")
    execution.cancel()

    logging.info("---- > request_id " + str(execution.request_id()))
    logging.info("---- > progress " + str(execution.progress()))
    logging.info("---- > status " + str(execution.status()))
    logging.info("---- > result " + str(execution.data()))
    logging.info("---- > messages " + str(execution.messages()))

    '''
    #wrapper.state("aaa")

    logging.info(f"Execute synced")
    response = wrapper.execute(input)
    logging.info(f"Response {response}")

    logging.info(f"Execute async")
    response = wrapper.execute_async(input)

    logging.info(f"Get async response")
    logging.info(f"Response {response.result()}")

    # Check exit is ok
    time.sleep(2)
    '''
    wrapper.__del__()



from openfabric_pysdk.utility import LoaderUtil

# Execution callback function
execution_callback_function = LoaderUtil.get_function("main_callback")

# Config callback function
config_callback_function = LoaderUtil.get_function("config_callback")

# Config callback function
suspend_callback_function = LoaderUtil.get_function("suspend_callback")
